#!/bin/sh

# AUTHOR         : Andreas Bollhalder (bolle@geodb.org)
# DATE           : 17. November 2007
# REVISION	     : 14. Mai 2010 by Andreas Bollhalder
# VERSION        : 0.31 alpha 4
# LANGUAGE       : Linux shell script (/bin/sh)
# DEPENDENCIES
# BINARIES       : 'util-linux', 'bridge-utils', 'uml-utilities'
# KERNEL MODULES : 'tun' or 'bridge'

# SYNOPSIS:
#  This script set up a TAP adapter to be used by DOSEMU and configures PC/GEOS
#  to use the right IP address, domain name servers and HTTP proxy server.

# CALLED BY:
#  Linux super user 'root' by hand or system.

# ARGUMENTS:
#  none
#  '-v' | '--version' | 'version'
#  '-c' | '--config' | 'config'
#  '-h' | '--help' | 'help'
#  '-u' | '--up' | 'start'
#  '-d' | '--down' | 'stop'
#  '-r' | '--restart' | 'restart'
#  '-l' | '--reload' | 'reload'
#  '-s' | '--status' | 'status'
# ['--complete']
# ['--force']
# ['--verbose']

# RETURN VALUES:
#  0	if successful
#  1	if error

# SIDE EFFECTS:
#  None.

# STRATEGY:
#  1.	Define the user variables.
#  2.	Define the global variables.
#  3.	Define the helper functions.
#  4.	Define the worker functions.
#  5.	Define the main functions.
#  6.	Execute the selected command.

# TODOS:
#  - Check TAP and BRIDGE kernel support
#  - Improve advices and error messages
#  - Improve help
#  - Print commands in verbose mode
#  - Check all needed entries in GEOS.INI
#  - Implement '--auto' modus:
#    * TAP and GEOS active -> reload
#    * TAP and GEOS inactive -> start
#    * TAP active, GEOS inactive -> restart
#    * TAP inactive, GEOS active -> none
#  - Activate removing the firewall NAT routing for the host interface
#  - Check if invoked by a user with sufficient rights instead of root
#  - Check DOSEMU configuration
#  - Start DOSEMU
#  - Cleanups


# ============================================================================
#  1.	Define the user variables.
#     Please edit this variables according your installation.
# ----------------------------------------------------------------------------

# Set this to 0 after adjusting the user variables.
U_SCRIPT_LOCKED=0

# The interface name of the bridge device. It must be unique on the system.
U_BRI_INTERFACE="br0"

# Set this to 0 if you must load a kernel module for the TAP adapter, otherwise
# set it to 1 if the TAP adapter is compiled into the kernel.
U_TAP_IN_KERNEL=1

# Set this to 0 if you must load a kernel module for the bridge, otherwise
# set it to 1 if the bridge support is compiled into the kernel.
U_BRI_IN_KERNEL=0

# The interface name of the TAP adapter. It must be the same as configured in
# the DOSEMU configuration file. It must start with 'tap' and must be unique
# on the system.
U_TAP_INTERFACE="tap0"

# The name of the unix user which will use the TAP adapter.
U_TAP_USER="andreas"

# Specify the path of the GEOS root directory.
U_GEOS_ROOT="/home/$U_TAP_USER/Programs/bbe413en/drive_c/ensemble"

# Specify the 'GEOS.INI' file which will be configured.
U_GEOS_INI="$U_GEOS_ROOT/geos.ini"

# Specify the 'GEOS_ACT.IVE' file which will be checked.
U_GEOS_ACTIVE="$U_GEOS_ROOT/geos_act.ive"


# ============================================================================
#  2.	Define the global variables.
#     !!! Only change this variables if you know what you are doing. !!!
# ----------------------------------------------------------------------------

# Script command and options
G_ARG_CMD=""
G_ARG_OPT_COMPLETE=0
G_ARG_OPT_FORCE=0
G_ARG_OPT_VERBOSE=0

# Script version and date
G_SCRIPT_VERSION="0.31 alpha 4"
G_SCRIPT_DATE="20100514"

# Space seperated list of paths to test for the existence of the 'which'
# binary.
G_WHICH_PATHS="/usr/bin /bin /sbin /usr/sbin"

# Space seperated list of required Linux binaries.
G_REQ_LINUX_BINS="awk expr grep ifconfig less lsmod modinfo modprobe passwd route sed seq"
# Space seperated list of required bridge binaries.
G_REQ_BRIDGE_BINS="brctl"
# Space seperated list of required UML binaries.
G_REQ_UML_BINS="tunctl"
# Space seperated list of required iptables binaries.
G_REQ_IPTABLES_BINS="iptables"

# Space seperated list of required proc files.
G_REQ_PROC_FILES="/proc/sys/net/ipv4/ip_forward"

# The kernel module which provides the bridge support.
G_BRI_KERNEL_MOD="bridge"
# The kernel module which provides the TAP adapter.
G_TAP_KERNEL_MOD="tun"

# The proc file which controls the ip forwarding.
G_PROC_NET_IP_FORWARD="/proc/sys/net/ipv4/ip_forward"

# The active network interface which will be bridged with the TAP adapter.
G_NET_INTERFACE=""
# The current IP address of the active netwerk interface.
G_NET_IP_ADDR=""

# Space seperated list with IP addresses of the domain name servers.
G_SYS_DNS_SERVERS=""

# The HTTP proxy server.
G_SYS_HTTP_PROXY=""

# The IP address of the interface of the bridge device.
G_BRI_IP_NET_12="192.168"
G_BRI_IP_ADDR=""
G_BRI_IP_ADDR_4="254"
G_BRI_IP_MASK="255.255.255.0"

# The version of GEOS.
G_GEOS_VERSION=""
G_REQ_GEOS_VERSION=4

# The IP address used by GEOS.
G_GEOS_IP_ADDR=""
G_GEOS_IP_ADDR_4="200"
G_GEOS_DNS1=""
G_GEOS_DNS2=""
G_GEOS_HTTP_PROXY=""

# The 'GEOS.INI' entries which the script will use or update.
G_INI_SEC_SYS="system"
G_INI_KEY_SYS_VER="version"

G_INI_SEC_AP2="accessPoint0002"
G_INI_KEY_AP2_NAME="name"
G_INI_KEY_AP2_IP_ADDR="ipaddr"
G_INI_KEY_AP2_IP_GATE="ipgate"
G_INI_KEY_AP2_IP_MASK="ipmask"
G_INI_KEY_AP2_DNS1="dns1"
G_INI_KEY_AP2_DNS2="dns2"

G_INI_SEC_WEB="HTMLView"
G_INI_KEY_WEB_WWW_PROXY="WWWProxy"


# ============================================================================
#  3.	Define the helper functions.
# ----------------------------------------------------------------------------

#
# FUNCTION: handleError
#
# DESCRIPTION:
#  This function exits the script of the option '--force' is not used.
#
handleError()
{
  if [ $G_ARG_OPT_FORCE -eq 0 ] ;
  then
    exit 1
  fi
}

#
# FUNCTION: printVerbose $1
#  $1		: The text to print. If 0, print '[OK]'. If 1, print '[FAILED]'.
#
# DESCRIPTION:
#  This function print verbose information if the option '--verbose' is used.
#
printVerbose()
{
  if [ $G_ARG_OPT_VERBOSE -eq 1 ] ;
  then
    case $1 in
    0 )
      echo "[OK]"
    ;;
    1 )
      echo "[FAILED]"
    ;;
    2 )
      echo "[SKIPPED]"
    ;;
    * )
      echo -n $1
    ;;
    esac
  fi
  return
}


# ============================================================================
#  4.	Define the worker functions.
# ----------------------------------------------------------------------------

#
# FUNCTION: checkScriptLocked
#
# DESCRIPTION:
#  This function checks if the skript is unlocked.
#
checkScriptLocked()
{
  printVerbose "Checking if script is locked ... "
  if [ $U_SCRIPT_LOCKED -eq 1 ] ; then
    printVerbose 1
    echo ""
    echo "The script is locked !!!"
    echo "Please edit the user variables in this script and then set"
    echo "'U_SCRIPT_LOCKED' to '0'."
    echo ""
    # Exit the script even if the option '--force' is used.
    exit 1
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: checkInvocedByRoot
#
# DESCRIPTION:
#  This function checks if the skript is invoced by root.
#
checkInvocedByRoot()
{
  printVerbose "Checking if script is invoced by root ... "
  if ! [ "$USER" = "root" ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The script must be invoced by the user 'root' !!!"
    echo ""
    # Exit the script even if the option '--force' is used.
    exit 1
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: checkWhichBinary $1
#  $1		: Space seperated list of absolute paths to check.
#
# DESCRIPTION:
#  This function checks if the 'which' binary exist on the system.
#
checkWhichBinary()
{
  printVerbose "Looking up 'which' binary ... "
  for L_PATH in $1
  do
    if [ -f "$L_PATH/which" ] ;
    then
      printVerbose 0
      return
    fi
  done
  printVerbose 1
  echo ""
  echo "  ERROR - Could not find the binary 'which' !!!"
  echo ""
  echo "Please install it to your system or check that it can be found under"
  echo "one of the following paths:"
  echo ""
  for L_PATH in $1
  do
    echo $L_PATH
  done
  echo ""
  # Exit the script even if the option '--force' is used.
  exit 1
}

#
# FUNCTION: checkLinuxBinaries $1
#  $1		: Space seperated list of Linux binaries to check.
#
# DESCRIPTION:
#  This function checks a list of Linux binaries if they exist on the system.
#
checkLinuxBinaries()
{
  printVerbose "Looking up Linux binaries ... "
  L_ERR=0
  for L_BINARY in $1
  do
    L_RET=`which $L_BINARY`
    if ! [ $? -eq 0 ] ;
    then
      printVerbose 1
      echo ""
      echo "  ERROR - Could not find a Linux binary !!!"
      echo "  VALUE : '$L_BINARY'"
      L_ERR=1
    fi
  done
  if [ $L_ERR -ne 0 ] ;
  then
    echo ""
    echo "One or more required Linux binaries could not be found."
    echo "Please install the appropriate packages to your system."
    echo ""
    # Exit the script even if the option '--force' is used.
    exit 1
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: checkPackageBinaries $1
#  $1		: Space seperated list of package binaries to check.
#  $2		: Package name.
#  $3		: Package RPM name.
#
# DESCRIPTION:
#  This function checks a list of package binaries if they exist on the system.
#
checkPackageBinaries()
{
  printVerbose "Looking up $2 binaries ... "
  L_ERR=0
  for L_BINARY in $1
  do
    L_RET=`which $L_BINARY`
    if ! [ $? -eq 0 ] ;
    then
      printVerbose 1
      echo ""
      echo "  ERROR - Could not find a $2 binary !!!"
      echo "  VALUE : '$L_BINARY'"
      L_ERR=1
    fi
  done
  if [ $L_ERR -ne 0 ] ;
  then
    echo ""
    echo "One or more required $2 binaries could not be found."
    echo "The package '$3' must be installed on your system."
    echo ""
    # Exit the script even if the option '--force' is used.
    exit 1
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: checkProcFiles $1
#  $1		: Space seperated list of proc files to check.
#
# DESCRIPTION:
#  This function checks a list of proc files if they exist on the system.
#
checkProcFiles()
{
  printVerbose "Checking proc files ... "
  L_ERR=0
  for L_FILE in $1
  do
    if ! [ -f $L_FILE ] ;
    then
      printVerbose 1
      echo ""
      echo "  ERROR - Could not find a proc file !!!"
      echo "  VALUE : '$L_FILE'"
      L_ERR=1
    fi
  done
  if [ $L_ERR -ne 0 ] ;
  then
    echo ""
    echo "One or more required proc file could not be found."
    echo "Please check your kernel and tools configuration of your system."
    echo ""
    # Exit the script even if the option '--force' is used.
    exit 1
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: checkTapUser
#
# DESCRIPTION:
#  This function checks if the user of the TAP adapter exist.
#
checkTapUser()
{
  printVerbose "Checking user of TAP adapter ... "
  L_RET=`passwd -S "$U_TAP_USER"`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The user of the TAP adapter does not exist !!!"
    echo "  VALUE : '$U_TAP_USER'"
    echo ""
    echo "Please check the script variable 'U_TAP_USER'."
    echo ""
    # Exit the script even if the option '--force' is used.
    exit 1
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: checkTapName
#
# DESCRIPTION:
#  This function checks if the interface name of the TAP adapter starts with 'tap'.
#
checkTapName()
{
  printVerbose "Checking interface name of TAP adapter ... "
  L_RET=`echo $U_TAP_INTERFACE|awk '{ print substr($1,1,3) }'`
  if ! [ "$L_RET" = "tap" ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The interface name of the TAP adapter does not start with 'tap' !!!"
    echo "  VALUE : '$U_TAP_INTERFACE' $L_RET"
    echo ""
    echo "Please check the script variable 'U_TAP_INTERFACE'."
    echo ""
    # Exit the script even if the option '--force' is used.
    exit 1
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: checkGeosIniFile
#
# DESCRIPTION:
#  This function checks if the 'GEOS.INI' file exist and that GEOS has the
#  version '4'.
#
checkGeosIniFile()
{
  # Check if the 'GEOS.INI' file exist.
  printVerbose "Checking existance of 'GEOS.INI' file ... "
  if ! [ -f "$U_GEOS_INI" ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The 'GEOS.INI' file could not be found !!!"
    echo "  VALUE : '$U_GEOS_INI'"
    echo ""
    echo "Please check the script variable 'U_GEOS_INI'."
    echo ""
    # Exit the script even if the option '--force' is used.
    exit 1
  else
    printVerbose 0
  fi
  # Get the GEOS version number from the 'GEOS.INI'.
  printVerbose "Retrieving 'GEOS.INI' version number ... "
  G_GEOS_VERSION=`awk -F "=" ' \
  $1 ~ /^\['$G_INI_SEC_SYS'\]/, $1 ~ /^'$G_INI_KEY_SYS_VER'/ || /^\[.*$/ && ! /^\['$G_INI_SEC_SYS'\]/ \
  { \
    sub(/^[ \t]+/,"",$1); \
    sub(/[ \t]+$/,"",$1); \
    if ( $1 == "'$G_INI_KEY_SYS_VER'" ) \
    { \
      sub(/^[ \t]+/,"",$2); \
      sub(/[ \t]+$/,"",$2); \
      print $2; \
      exit \
    } \
  }' "$U_GEOS_INI"|sed 's/\r//'`
  # Check if the version number exist.
  if [ -z "$G_GEOS_VERSION" ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The 'GEOS.INI' does not contain a version number !!!"
    echo "  VALUE : '$G_GEOS_VERSION'"
    echo ""
    echo "Please verify your 'GEOS.INI' or install a correct version."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  # Check that the major version number is 4.
  printVerbose "Checking 'GEOS.INI' version number ... "
  if ! [ `expr substr $G_GEOS_VERSION 1 1` -eq $G_REQ_GEOS_VERSION ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The 'GEOS.INI' major version is not equal to '$G_REQ_GEOS_VERSION' !!!"
    echo "  VALUE : '$G_GEOS_VERSION'"
    echo ""
    echo "Only Breadbox Ensemble, Global PC or NewDeal Office 2000 are supported."
    echo "Please check your 'GEOS.INI' or install a correct version."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: checkGeosActiveFile
#
# DESCRIPTION:
#  This function checks if the 'GEOS_ACT.IVE' file does exist.
#
checkGeosActiveFile()
{
  printVerbose "Checking existance of 'GEOS_ACT.IVE' file ... "
  if [ -f "$U_GEOS_ACTIVE" ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The 'GEOS_ACT.IVE' file has been found !!!"
    echo "  VALUE : '$U_GEOS_ACTIVE'"
    echo ""
    echo "Please verify that GEOS is not running, otherwise remove the file."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: checkBridgeKernelModule
#
# DESCRIPTION:
#  This function checks if the kernel module for the bridge exist.
#
checkBridgeKernelModule()
{
  printVerbose "Checking kernel module for bridge ... "
  # Check for bridge in-kernel module.
  if [ $U_BRI_IN_KERNEL -eq 1 ] ;
  then
    printVerbose 2
    return 
  fi
  # Check for the bridge kernel module.
  L_RET=`modinfo $G_BRI_KERNEL_MOD`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - Could not find the kernel module for the bridge !!!"
    echo "  VALUE : '$G_BRI_KERNEL_MOD'"
    echo ""
    echo "The kernel module for the bridge could not be found."
    echo "Please check your kernel and modules of your system."
    echo ""
    handleError
    return
  fi
  printVerbose 0
  return 
}

#
# FUNCTION: checkTapKernelModule
#
# DESCRIPTION:
#  This function checks if the kernel module for the TAP adapter exist.
#
checkTapKernelModule()
{
  printVerbose "Checking kernel module for TAP adapter ... "
  # Check for TAP in-kernel module.
  if [ $U_TAP_IN_KERNEL -eq 1 ] ;
  then
    printVerbose 2
    return 
  fi
  # Check for the TAP kernel module.
  L_RET=`modinfo $G_TAP_KERNEL_MOD`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - Could not find the kernel module for the TAP adapter !!!"
    echo "  VALUE : '$G_TAP_KERNEL_MOD'"
    echo ""
    echo "The kernel module for the TAP adapter could not be found."
    echo "Please check your kernel and modules of your system."
    echo ""
    handleError
    return
  fi
  printVerbose 0
  return 
}

#
# FUNCTION: findActiveNetworkInterface
#
# DESCRIPTION:
#  This function search the active network interface of the system.
#  route | awk '/^default / { print $NF }'
#
findActiveNetworkInterface()
{
  printVerbose "Looking up active host network interface ... "
  G_NET_INTERFACE=`LANG=C route -n|awk '/^0.0.0.0 / { print $NF }'`
  if [ -z "$G_NET_INTERFACE" ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - An active host network interface could not be found !!!"
    echo "  VALUE : '$G_NET_INTERFACE'"
    echo ""
    echo "Please check your network setup."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: findActiveIpAddress
#
# DESCRIPTION:
#  This function search the active IP address of the system.
#
findActiveIpAddress()
{
  printVerbose "Looking up IP address of active host network interface ... "
  G_NET_IP_ADDR=`LANG=C ifconfig $G_NET_INTERFACE|awk '/inet / { print $2 }'|sed "s/addr://"`
  if [ -z "$G_NET_IP_ADDR" ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - An IP address could not be found !!!"
    echo "  VALUE : '$G_NET_IP_ADDR'"
    echo ""
    echo "Please check your network setup."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: findFreeBridgeNetwork
#
# DESCRIPTION:
#  This function search for a free network in the system.
#
findFreeBridgeNetwork()
{
  printVerbose "Searching free network for bridge device ... "
  L_IP_ADDRS=`LANG=C ifconfig|awk 'BEGIN { ORS=" " } /inet / { print $2 }'|sed 's/addr://g; s/ $//'`
  for L_IP_NET in `seq 0 255`
  do
    L_USED=0
    for L_IP_IN_USE in $L_IP_ADDRS
    do
      if [ `echo $L_IP_IN_USE|grep $G_BRI_IP_NET_12.$L_IP_NET` ] ;
      then
        L_USED=1
        break
      fi
    done
    if [ $L_USED -eq 0 ] ;
    then
      G_BRI_IP_ADDR=$G_BRI_IP_NET_12.$L_IP_NET.$G_BRI_IP_ADDR_4
      G_GEOS_IP_ADDR=$G_BRI_IP_NET_12.$L_IP_NET.$G_GEOS_IP_ADDR_4
      printVerbose 0
      return
    fi
  done
  printVerbose 1
  echo ""
  echo "  ERROR - A free network for the bridge device could not be found !!!"
  echo "  VALUE : '$G_BRI_IP_NET_12.[0..255]'"
  echo ""
  echo "Please check your network setup."
  echo ""
  handleError
  return
}

#
# FUNCTION: findActiveBridgeNetwork
#
# DESCRIPTION:
#  This function search the active network of the bridge device.
#
findActiveBridgeNetwork()
{
  printVerbose "Looking up active network of bridge device ... "
  G_BRI_IP_ADDR=`LANG=C ifconfig $U_BRI_INTERFACE|awk '/inet / { print $2 }'|sed "s/addr://"`
  if [ -z "$G_BRI_IP_ADDR" ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The active network of the bridge device could not be found !!!"
    echo "  VALUE : '$G_BRI_IP_ADDR'"
    echo ""
    echo "Please check your network setup."
    echo ""
    handleError
    return
  fi
  L_IP_NET=`echo $G_BRI_IP_ADDR|awk 'BEGIN { FS = "." } ; { print $3 }'`
  G_GEOS_IP_ADDR=$G_BRI_IP_NET_12.$L_IP_NET.$G_GEOS_IP_ADDR_4
  printVerbose 0
  return
}

#
# FUNCTION: findDomainNameServers
#
# DESCRIPTION:
#  This function search the domain name servers of the system.
#
findDomainNameServers()
{
  printVerbose "Looking up DNS servers ... "
  G_SYS_DNS_SERVERS=`awk 'BEGIN { ORS= " " } /^nameserver / { print $2 }' /etc/resolv.conf|sed 's/ $//'`
  if [ -z "$G_SYS_DNS_SERVERS" ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - A domain name server could not be found !!!"
    echo "  VALUE : '$G_SYS_DNS_SERVERS'"
    echo ""
    echo "Please check your '/etc/resolv.conf'."
    echo ""
    handleError
    return
  fi
  G_GEOS_DNS1=`echo $G_SYS_DNS_SERVERS|awk '{ print $1 }'`
  G_GEOS_DNS2=`echo $G_SYS_DNS_SERVERS|awk '{ print $2 }'`
  if [ -z "$G_GEOS_DNS2" ] ;
  then
    G_GEOS_DNS2=$G_GEOS_DNS1
  fi
  printVerbose 0
  return
}

#
# FUNCTION: findHttpProxyServer
#
# DESCRIPTION:
#  This function search the HTTP proxy server of the system.
#
findHttpProxyServer()
{
  printVerbose "Looking up HTTP proxy server ... "
  G_SYS_HTTP_PROXY=$http_proxy
  if [ -z "$G_SYS_HTTP_PROXY" ] ;
  then
    G_GEOS_HTTP_PROXY="{}"
  else
    G_GEOS_HTTP_PROXY=$G_SYS_HTTP_PROXY
  fi
  printVerbose 0
  return
}

#
# FUNCTION: loadBridgeKernelModule
#
# DESCRIPTION:
#  This function loads the kernel module for the bridge if not loaded.
#
loadBridgeKernelModule()
{
  printVerbose "Loading kernel module for bridge ... "
  # Check for bridge in-kernel module.
  if [ $U_BRI_IN_KERNEL -eq 1 ] ;
  then
    printVerbose 2
    return 
  fi
  # Check if the kernel module for the bridge is already loaded.
  L_RET=`lsmod|awk '/'$G_BRI_KERNEL_MOD' / { print $1 }'`
  if [ -z "$L_RET" ] ;
  then
    # Load the kernel module for the bridge.
    L_RET=`modprobe $G_BRI_KERNEL_MOD`
    if ! [ $? -eq 0 ] ; then
      printVerbose 1
      echo ""
      echo "  ERROR - The kernel module for the bridge could not be loaded."
      echo "  VALUE : '$G_BRI_KERNEL_MOD'"
      echo ""
      echo "Please check the kernel and modules."
      echo ""
      handleError
      return
    fi
  else
    printVerbose 2
    return 
  fi
  printVerbose 0
  return
}

#
# FUNCTION: loadTapKernelModule
#
# DESCRIPTION:
#  This function loads the kernel module for the TAP adapter if not loaded.
#
loadTapKernelModule()
{
  printVerbose "Loading kernel module for TAP adapter ... "
  # Check for TAP in-kernel module.
  if [ $U_TAP_IN_KERNEL -eq 1 ] ;
  then
    printVerbose 2
    return 
  fi
  # Check if the kernel module for the TAP adapter is already loaded.
  L_RET=`lsmod|awk '/'$G_TAP_KERNEL_MOD' / { print $1 }'`
  if [ -z "$L_RET" ] ;
  then
    # Load the kernel module for the TAP adapter.
    L_RET=`modprobe $G_TAP_KERNEL_MOD`
    if ! [ $? -eq 0 ] ; then
      printVerbose 1
      echo ""
      echo "  ERROR - The kernel module for the TAP adapter could not be loaded."
      echo "  VALUE : '$G_TAP_KERNEL_MOD'"
      echo ""
      echo "Please check the kernel and modules."
      echo ""
      handleError
      return
    fi
  else
    printVerbose 2
    return 
  fi
  printVerbose 0
  return
}

#
# FUNCTION: unloadBridgeKernelModule
#
# DESCRIPTION:
#  This function removes the kernel modules.
#
unloadBridgeKernelModule()
{
  printVerbose "Unloading kernel module for bridge ... "
  # Check for bridge in-kernel module.
  if [ $U_BRI_IN_KERNEL -eq 1 ] ;
  then
    printVerbose 2
    return 
  fi
  # Unload the kernel module for the bridge.
  L_RET=`rmmod "$G_BRI_KERNEL_MOD"`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The kernel module for the bridge could not be unloaded."
    echo "  VALUE : '$G_BRI_KERNEL_MOD'"
    echo ""
    echo "Please check the kernel and modules."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: unloadTapKernelModule
#
# DESCRIPTION:
#  This function removes the kernel modules.
#
unloadTapKernelModule()
{
  printVerbose "Unloading kernel module for TAP adapter ... "
  # Check for TAP in-kernel module.
  if [ $U_TAP_IN_KERNEL -eq 1 ] ;
  then
    printVerbose 2
    return 
  fi
  # Unload the kernel module for the TAP adapter.
  L_RET=`rmmod "$G_TAP_KERNEL_MOD"`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The kernel module for the TAP adapter could not be unloaded."
    echo "  VALUE : '$G_TAP_KERNEL_MOD'"
    echo ""
    echo "Please check the kernel and modules."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: enableIpForwarding
#
# DESCRIPTION:
#  This function enables the IP forwarding of the kernel.
#
enableIpForwarding()
{
  printVerbose "Enabling IP forwarding ... "
  L_RET=`echo 1 > "$G_PROC_NET_IP_FORWARD"`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The IP forwarding of the kernel could not be enabled."
    echo "  VALUE : '$G_PROC_NET_IP_FORWARD'"
    echo ""
    echo "Please check the kernel and modules."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: disableIpForwarding
#
# DESCRIPTION:
#  This function disables the IP forwarding of the kernel.
#
disableIpForwarding()
{
  printVerbose "Disabling IP forwarding ... "
  L_RET=`echo 0 > "$G_PROC_NET_IP_FORWARD"`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The IP forwarding of the kernel could not be disabled."
    echo "  VALUE : '$G_PROC_NET_IP_FORWARD'"
    echo ""
    echo "Please check the network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: setupNetworkBridge
#
# DESCRIPTION:
#  This function create the bridge device with STP disabled and brings up its
#  interface.
#
setupNetworkBridge()
{
  # Create the bridge device
  printVerbose "Creating bridge device ... "
  L_RET=`brctl addbr "$U_BRI_INTERFACE"`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The bridge device could not be created."
    echo "  VALUE : '$U_BRI_INTERFACE'"
    echo ""
    echo "Please check the network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  # Disable the spanning tree protocol on the bridge device
  printVerbose "Disabling STP protocol on bridge device ... "
  L_RET=`brctl stp "$U_BRI_INTERFACE" off`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The STP protocol on the bridge device could not be disabled."
    echo "  VALUE : '$U_BRI_INTERFACE'"
    echo ""
    echo "Please check the network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  # Set up the interface of the bridge device
  printVerbose "Bringing up interface of bridge device ... "
  L_RET=`ifconfig "$U_BRI_INTERFACE" "$G_BRI_IP_ADDR" netmask "$G_BRI_IP_MASK" up`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The interface of the bridge device could not be brought up."
    echo "  VALUE : '$U_BRI_INTERFACE'"
    echo ""
    echo "Please check the network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: removeNetworkBridge
#
# DESCRIPTION:
#  This function shuts down the bridge device and then deletes it.
#
removeNetworkBridge()
{
  # Shut down the interface of the bridge device
  printVerbose "Shutting down interface of bridge device ... "
  L_RET=`ifconfig "$U_BRI_INTERFACE" down`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The interface of the bridge device could not be deactivated."
    echo "  VALUE : '$U_BRI_INTERFACE'"
    echo ""
    echo "Please check the network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  # Delete the bridge device
  printVerbose "Deleting bridge device ... "
  L_RET=`brctl delbr "$U_BRI_INTERFACE"`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The bridge device could not be deleted."
    echo "  VALUE : '$U_BRI_INTERFACE'"
    echo ""
    echo "Please check the network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: setupTapAdapter
#
# DESCRIPTION:
#  This function create the TAP adapter, adds it to the bridge device and
#  brings up its interface.
#
setupTapAdapter()
{
  # Create the TAP adapter
  printVerbose "Creating TAP adapter ... "
  L_RET=`tunctl -b -u "$U_TAP_USER" -t "$U_TAP_INTERFACE"`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The TAP adapter could not be created."
    echo "  VALUE : '$U_TAP_USER'"
    echo ""
    echo "Please check the user or network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  # Add the TAP adapter to the bridge device
  printVerbose "Adding TAP adapter to bridge device ... "
  L_RET=`brctl addif "$U_BRI_INTERFACE" "$U_TAP_INTERFACE"`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The TAP adapter could not be added to the bridge device."
    echo "  VALUE : '$U_TAP_INTERFACE'"
    echo ""
    echo "Please check the network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  # Set up the interface of the TAP adapter
  printVerbose "Bringing up interface of TAP adapter ... "
  L_RET=`ifconfig "$U_TAP_INTERFACE" up 0.0.0.0 promisc`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The interface of the TAP adapter could not be brought up."
    echo "  VALUE : '$U_TAP_INTERFACE'"
    echo ""
    echo "Please check the network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: removeTapAdapter
#
# DESCRIPTION:
#  This function shuts down the TAP adapter, removes it from the bridge device
#  and then deletes it.
#
removeTapAdapter()
{
  # Shut down the interface of the TAP adapter
  printVerbose "Shutting down interface of TAP adapter ... "
  L_RET=`ifconfig "$U_TAP_INTERFACE" down`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The interface of the TAP adapter could not be shut down."
    echo "  VALUE : '$U_TAP_INTERFACE'"
    echo ""
    echo "Please check the network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  # Remove the TAP adapter from the bridge device
  printVerbose "Removing TAP adapter from bridge device ... "
  L_RET=`brctl delif "$U_BRI_INTERFACE" "$U_TAP_INTERFACE"`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The TAP adapter could not be removed from the bridge device."
    echo "  VALUE : '$U_TAP_INTERFACE'"
    echo ""
    echo "Please check the network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  # Delete the TAP adapter
  printVerbose "Deleting TAP adapter ... "
  L_RET=`tunctl -d "$U_TAP_INTERFACE"`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The TAP adapter could not be deleted."
    echo "  VALUE : '$U_TAP_INTERFACE'"
    echo ""
    echo "Please check the network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: setupIpTables
#
# DESCRIPTION:
#  This function set up iptables to allow traffic between the bridge and the
#  TAP adapter and configures the masquerading (NAT).
#
setupIpTables()
{
  # Add the firewall rules for the TAP adapter
  printVerbose "Adding firewall rules for TAP adapter ... "
  L_RET=`iptables -A INPUT -i "$U_TAP_INTERFACE" -j ACCEPT`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The iptables rules for the TAP adapter could not be added."
    echo "  VALUE : '$U_TAP_INTERFACE'"
    echo ""
    echo "Please check the network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  # Add the firewall rules for the bridge device
  printVerbose "Adding firewall rules for bridge device ... "
  L_RET=`iptables -A INPUT -i "$U_BRI_INTERFACE" -j ACCEPT`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The iptables rules for the bridge could not be added."
    echo "  VALUE : '$U_BRI_INTERFACE'"
    echo ""
    echo "Please check the network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  # Add the firewall forwarding to the bridge device
  printVerbose "Adding firewall forwarding for for bridge device ... "
  L_RET=`iptables -A FORWARD -i "$U_BRI_INTERFACE" -j ACCEPT`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The iptables forwarding for the bridge device could not be added."
    echo "  VALUE : '$U_BRI_INTERFACE'"
    echo ""
    echo "Please check the network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  # Add the firewall NAT routing for the host interface
  printVerbose "Adding firewall NAT routing for host interface ... "
  L_RET=`iptables -t nat -A POSTROUTING -o "$G_NET_INTERFACE" -j MASQUERADE`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The iptables post routing for the host interface could not be added."
    echo "  VALUE : '$G_NET_INTERFACE'"
    echo ""
    echo "Please check the network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: cleanupIpTables
#
# DESCRIPTION:
#  This function removes from the iptables the masquerading (NAT) and the rules
#  for the traffic between the bridge and TAP adapter.
#
cleanupIpTables()
{
  # Remove the firewall forwarding from the bridge device
  printVerbose "Removing firewall forwarding from bridge device ... "
  L_RET=`iptables -D FORWARD -i "$U_BRI_INTERFACE" -j ACCEPT`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The iptables forwarding for the bridge device could not be removed."
    echo "  VALUE : '$U_BRI_INTERFACE'"
    echo ""
    echo "Please check the network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  # Remove the firewall rules for the TAP adapter
  printVerbose "Removing firewall rules for TAP adapter ... "
  L_RET=`iptables -D INPUT -i "$U_TAP_INTERFACE" -j ACCEPT`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The iptables rules for the TAP adapter could not be removed."
    echo "  VALUE : '$U_TAP_INTERFACE'"
    echo ""
    echo "Please check the network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  # Remove the firewall rules for the bridge device
  printVerbose "Removing firewall rules for bridge device ... "
  L_RET=`iptables -D INPUT -i "$U_BRI_INTERFACE" -j ACCEPT`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The iptables rules for the bridge device could not be removed."
    echo "  VALUE : '$U_BRI_INTERFACE'"
    echo ""
    echo "Please check the network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  # Remove the firewall NAT routing for the host interface
  printVerbose "Removing firewall NAT routing for host interface... "
#  L_RET=`iptables -t nat -D POSTROUTING -o "$G_NET_INTERFACE" -j MASQUERADE`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The iptables post routing for the host interface could not be removed."
    echo "  VALUE : '$G_NET_INTERFACE'"
    echo ""
    echo "Please check the network configuration."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  return
}

#
# FUNCTION: updateGeosIniFile
#
# DESCRIPTION:
#  This function updates the 'GEOS.INI' file.
#
updateGeosIniFile()
{
  # 
  printVerbose "Initializing 'GEOS.INI' access point section ... "
  L_RET=`grep "$G_INI_SEC_AP2" "$U_GEOS_INI"`
  if ! [ $? -eq 0 ] ;
  then
    L_RET=`echo -n "\r[$G_INI_SEC_AP2]\rname = DOSEMU-TAP BR\rnetType = 1\rdefault = 1\ripaddr = 192.168.0.200\ripgate = 192.168.0.254\ripmask = 255.255.255.0\rdns1 = \rdns2 = \rphone = 0\rpromptSecret = 0\rlocalDialingOptions = 0" >> "$U_GEOS_INI"`
    if ! [ $? -eq 0 ] ;
    then
      printVerbose 1
      echo ""
      echo "  ERROR - The 'GEOS.INI' file could not be initialized correctly."
      echo "  VALUE : '$G_INI_SEC_AP2'"
      echo ""
      echo "Please check the content of the file."
      echo ""
      handleError
    else
      printVerbose 0
    fi
  else
    printVerbose 0
  fi
  # 
  printVerbose "Initializing 'GEOS.INI' web section ... "
  L_RET=`grep "$G_INI_SEC_WEB" "$U_GEOS_INI"`
  if ! [ $? -eq 0 ] ;
  then
    L_RET=`echo "\r[$G_INI_SEC_WEB]\r" >> "$U_GEOS_INI"`
    if ! [ $? -eq 0 ] ;
    then
      printVerbose 1
      echo ""
      echo "  ERROR - The 'GEOS.INI' file could not be initialized correctly."
      echo "  VALUE : '$G_INI_SEC_WEB'"
      echo ""
      echo "Please check the content of the file."
      echo ""
      handleError
    else
      printVerbose 0
    fi
  else
    printVerbose 0
  fi
  # 
  printVerbose "Updating 'GEOS.INI' access point section ... "
  L_RET=`sed -i " \
  /^[[:space:]]*\[$G_INI_SEC_AP2\]/,/\[/ \
  { \
    s/^[[:space:]]*\[$G_INI_SEC_AP2\].*$/\[$G_INI_SEC_AP2\]\r/; \
    s/^[[:space:]]*$G_INI_KEY_AP2_NAME.*$/$G_INI_KEY_AP2_NAME = Linux $U_TAP_INTERFACE-$U_BRI_INTERFACE\r/; \
    s/^[[:space:]]*$G_INI_KEY_AP2_IP_ADDR.*$/$G_INI_KEY_AP2_IP_ADDR = $G_GEOS_IP_ADDR\r/; \
    s/^[[:space:]]*$G_INI_KEY_AP2_IP_MASK.*$/$G_INI_KEY_AP2_IP_MASK = $G_BRI_IP_MASK\r/; \
    s/^[[:space:]]*$G_INI_KEY_AP2_IP_GATE.*$/$G_INI_KEY_AP2_IP_GATE = $G_BRI_IP_ADDR\r/; \
    s/^[[:space:]]*$G_INI_KEY_AP2_DNS1.*$/$G_INI_KEY_AP2_DNS1 = $G_GEOS_DNS1\r/; \
    s/^[[:space:]]*$G_INI_KEY_AP2_DNS2.*$/$G_INI_KEY_AP2_DNS2 = $G_GEOS_DNS2\r/ \
  }" "$U_GEOS_INI"`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The 'GEOS.INI' file could not be updated correctly."
    echo "  VALUE : '$G_INI_SEC_AP2'"
    echo ""
    echo "Please check the content of the file."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  #
  printVerbose "Updating 'GEOS.INI' web section ... "
  # Attention: We use a ',' instead of '/' for the proxy.
  L_RET=`sed -i " \
  /^[[:space:]]*\[$G_INI_SEC_WEB\]/,/\[/ \
  { \
    s/^[[:space:]]*\[$G_INI_SEC_WEB\].*$/\[$G_INI_SEC_WEB\]\r/; \
    s,^[[:space:]]*$G_INI_KEY_WEB_WWW_PROXY.*$,$G_INI_KEY_WEB_WWW_PROXY = $G_GEOS_HTTP_PROXY\r, \
  }" "$U_GEOS_INI"`
  if ! [ $? -eq 0 ] ;
  then
    printVerbose 1
    echo ""
    echo "  ERROR - The 'GEOS.INI' file could not be updated correctly."
    echo "  VALUE : '$G_INI_SEC_WEB'"
    echo ""
    echo "Please check the content of the file."
    echo ""
    handleError
  else
    printVerbose 0
  fi
  return
}


# ============================================================================
#  5.	Define the main functions.
# ----------------------------------------------------------------------------

#
# FUNCTION: printScriptUsage
#
# DESCRIPTION:
#  This function prints the script usage.
#
printScriptUsage()
{
  echo ""
  echo "Usage: $(basename $0) command [options]"
  echo ""
  echo "Set up the network for PC/GEOS version '$G_REQ_GEOS_VERSION' and higher under DOSEMU."
  echo ""
  echo "The following commands are available:"
  echo "  -v, --version             Print the script version and date."
  echo "  -c, --config              Print the script configuration."
  echo "  -h, --help                Show the help."
  echo "  -u, --up, start         ! Set up the network."
  echo "  -d, --down, stop        ! Shut down the network."
  echo "  -r, --restart, restart  ! Restart the network."
  echo "  -l, --reload, reload    ! Reload the network."
  echo "  -s, --status, status      Print the running configuration."
  echo ""
  echo "All commands marked with an exclamation mark '!' must be run as the user 'root'."
  echo ""
  echo "The following options are available:"
  echo " [--complete]               Additionally remove the TAP kernel module and"
  echo "                            disable the the kernel IP forwarding."
  echo "                            Only valid in conjunction with '--down', '-d'"
  echo "                            or 'stop'."
  echo ""
  echo " [--force]                  Do not stop the script execution if an error"
  echo "                            occurs."
  echo ""
  echo " [--verbose]                Print verbose information for every step."
  echo ""
  echo "Use the option '-h' or '--help' for further information."
  echo ""
  echo "Report bugs to <bolle@geodb.org>. You must use the keywords '[GEOS]' and '[TAP]'"
  echo "in the subject of the mail."
  echo ""
  return
}

#
# FUNCTION: printScriptVersion
#
# DESCRIPTION:
#  This function prints the script version.
#
printScriptVersion()
{
  echo ""
  echo "Script version:"
  echo "  Version                     : '$G_SCRIPT_VERSION'"
  echo "  Date                        : '$G_SCRIPT_DATE'"
  echo ""
  return
}

#
# FUNCTION: printScriptConfig
#
# DESCRIPTION:
#  This function prints the script configuration.
#
printScriptConfig()
{
  echo ""
  echo "Script configuration:"
  if [ $U_SCRIPT_LOCKED -eq 0 ] ;
  then
    echo "  Script locked               : '$U_SCRIPT_LOCKED' (No)"
  else
    echo "  Script locked               : '$U_SCRIPT_LOCKED' (Yes)"
  fi
  if [ $U_BRI_IN_KERNEL -eq 0 ] ;
  then
    echo "  Bridge module in kernel     : '$U_BRI_IN_KERNEL' (No)"
  else
    echo "  Bridge module in kernel     : '$U_BRI_IN_KERNEL' (Yes)"
  fi
  echo "  Bridge interface            : '$U_BRI_INTERFACE'"
  if [ $U_TAP_IN_KERNEL -eq 0 ] ;
  then
    echo "  TAP module in kernel        : '$U_TAP_IN_KERNEL' (No)"
  else
    echo "  TAP module in kernel        : '$U_TAP_IN_KERNEL' (Yes)"
  fi
  echo "  TAP adapter interface       : '$U_TAP_INTERFACE'"
  echo "  TAP adapter user            : '$U_TAP_USER'"
  echo "  GEOS.INI file               : '$U_GEOS_INI'"
  echo ""
  return
}

#
# FUNCTION: printRunningConfig
#
# DESCRIPTION:
#  This function prints the running configuration.
#
printRunningConfig()
{
  echo ""
  echo "Running configuration:"
  echo "  Active network interface    : '$G_NET_INTERFACE'"
  echo "  Active IP address           : '$G_NET_IP_ADDR'"
  echo "  System domain name servers  : '$G_SYS_DNS_SERVERS'"
  echo "  System HTTP proxy           : '$G_SYS_HTTP_PROXY'"
  echo "  Bridge kernel module        : '$G_BRI_KERNEL_MOD'"
  echo "  Bridge interface            : '$U_BRI_INTERFACE'"
  echo "  Bridge IP address           : '$G_BRI_IP_ADDR'"
  echo "  Bridge IP mask              : '$G_BRI_IP_MASK'"
  echo "  TAP adapter kernel module   : '$G_TAP_KERNEL_MOD'"
  echo "  TAP adapter interface       : '$U_TAP_INTERFACE'"
  echo "  GEOS version                : '$G_GEOS_VERSION'"
  echo "  GEOS IP address             : '$G_GEOS_IP_ADDR'"
  echo "  GEOS IP mask                : '$G_BRI_IP_MASK'"
  echo "  GEOS IP gateway             : '$G_BRI_IP_ADDR'"
  echo "  GEOS 1st domain name server : '$G_GEOS_DNS1'"
  echo "  GEOS 2nd domain name server : '$G_GEOS_DNS2'"
  echo "  GEOS HTTP proxy             : '$G_GEOS_HTTP_PROXY'"
  echo ""
  return
}

#
# FUNCTION: printScriptHelp
#
# DESCRIPTION:
#  This function prints the script help.
#
printScriptHelp()
{
  echo ""
  echo "------------------------"
  echo "*** GEOS TAP Control ***"
  echo "------------------------"
  echo ""
  echo "!!! IMPORTANT NOTICE !!!"
  echo ""
  echo "This help information is displayed with the 'less' command. Press 'q' or use"
  echo "'Ctrl-c' to exit the help."
  echo "To scroll up and down, use the 'Up', 'Down' or 'PgUp', 'PgDown' keys. To get"
  echo "help for about the 'less' command, press 'h'."
  echo ""
  echo ""
  echo "TABLE OF CONTENT"
  echo ""
  echo "1.  Purpose"
  echo "2.  Usage"
  echo "3.  Requirements"
  echo "3.1 Linux Kernel"
  echo "3.2 Linux Tools"
  echo "3.3 Root User"
  echo "3.4 PC/GEOS Version"
  echo "4.  Configurations"
  echo "4.1 Configure this Script"
  echo "4.2 Configure DOSEMU"
  echo "4.3 Configure DOS"
  echo "4.4 Configure PC/GEOS"
  echo ""
  echo ""
  echo "1. PURPOSE"
  echo ""
  echo "This script set up a TAP adapter to be used by DOSEMU and configures PC/GEOS to"
  echo "use the right IP address, domain name servers and HTTP proxy server."
  echo ""
  echo "Only Breadbox Ensemble, Global PC or NewDeal Office 2000 are supported. Older"
  echo "versions do not provide a working 'EtherODI' or 'EtherPKT' driver."
  echo ""
  echo ""
  echo "2.  USAGE"
  echo ""
  echo "The script arguments are divided into a command and one or more space seperated"
  echo "options. You start the script as follows:"
  echo ""
  echo "    geos-tap-ctrl.sh command [options]"
  echo ""
  echo "If you start the script without any arguments, it will print out the usage."
  echo ""
  echo "The following commands are available:"
  echo ""
  echo "  -v, --version           Print the script version and date."
  echo ""
  echo "  -c, --config            Print the script configuration."
  echo ""
  echo "  -h, --help              Show the help."
  echo ""
  echo "  -u, --up, start         Set up the networking for DOSEMU. Create the bridge"
  echo "                          device, TAP adapter, set up the interfaces,"
  echo "                          routing and firewall rules. Configures the 'GEOS.INI'"
  echo "                          file for the network."
  echo ""
  echo "                          You can not start two times the DOSEMU networking"
  echo "                          without shutting it down before."
  echo ""
  echo "                          Make sure that PC/GEOS is not running when using this"
  echo "                          option."
  echo ""
  echo "                          You must run the script as the user 'root'."
  echo ""
  echo "  -d, --down, stop        Shut down the networking for DOSEMU. Removes the"
  echo "                          bridge device, TAP adapter, interfaces, routing and"
  echo "                          firewall rules."
  echo ""
  echo "                          Make sure that PC/GEOS is not running when using this"
  echo "                          option."
  echo ""
  echo "                          You must run the script as the user 'root'."
  echo ""
  echo "  -r, --restart, restart  This does the same as starting and stopping the"
  echo "                          networking for DOSEMU."
  echo ""
  echo "                          Use this if there has been a change in the host"
  echo "                          networking and the change would conflict with the"
  echo "                          current running configuration."
  echo ""
  echo "                          Normally, you should be able to use the reload"
  echo "                          option."
  echo ""
  echo "                          Make sure that PC/GEOS is not running when using this"
  echo "                          option."
  echo ""
  echo "                          You must run the script as the user 'root'."
  echo ""
  echo "  -l, --reload, reload    Reload the networking for DOSEMU. This will update"
  echo "                          the routing and firewall rules."
  echo ""
  echo "                          This option is especially useful, if there has been"
  echo "                          a change in the host networking."
  echo ""
  echo "                          Examples are:"
  echo ""
  echo "                           - Host interface change"
  echo "                             Host switched between Cable and Wireless"
  echo ""
  echo "                           - Host IP change"
  echo "                             Host changed IP address on its interface"
  echo ""
  echo "                          PC/GEOS can stay running when using this option."
  echo ""
  echo "                          You must run the script as the user 'root'."
  echo ""
  echo "  -s, --status, status    Print the running configuration."
  echo ""
  echo "The following options are available:"
  echo ""
  echo " [--complete]             Disable the kernel IP forwarding and remove the TAP"
  echo "                          kernel module."
  echo ""
  echo "                          This option is only valid with the command '--down',"
  echo "                          '-d' or 'stop'."
  echo ""
  echo "                          Make sure that nothing else is depending on this."
  echo ""
  echo " [--force]                Force the continuation of the script execution, even"
  echo "                          there has been an error. This is helpful to clean up"
  echo "                          if there is a problem, but can be harmful if your"
  echo "                          configuration is not correct and the script access"
  echo "                          someone others resources."
  echo ""
  echo "                          This option can be used with all commands."
  echo ""
  echo " [--verbose]              Print verbose information of the script steps."
  echo ""
  echo "                          This option can be used with all commands."
  echo ""
  echo ""
  echo "3.  REQUIREMENTS"
  echo ""
  echo "3.1 Linux Kernel"
  echo ""
  echo "You need a kernel with TUN and bridge support. This can be either compiled into"
  echo "the kernel itself or provided by a kernel module like 'tun' and/or 'bridge'."
  echo ""
  echo "The related kernel configurations are 'CONFIG_TUN' and 'CONFIG_BRIDGE'."
  echo ""
  echo "Sometimes, there are other dependend modules needed. For example, Puppy Linux"
  echo "uses the module 'bridge', which depends on the module 'llc'."
  echo ""
  echo "3.2 Linux Tools"
  echo ""
  echo "You need also the bridge and UML utilities and standard Linux tools installed."
  echo "The script checks most of the needed dependencies and tries to give you some"
  echo "hints if an error occures."
  echo ""
  echo "3.3 Root User"
  echo ""
  echo "You must be able to run this script as the user 'root'."
  echo ""
  echo "3.4 PC/GEOS Version"
  echo "You must have a legal copy of a PC/GEOS version '$G_REQ_GEOS_VERSION' or higher. This includes"
  echo "Breadbox Ensemble, Global PC and NewDeal Office 2000."
  echo ""
  echo "You must have either the 'EtherODI' or 'EtherPKT' driver for your version of"
  echo "PC/GEOS. The 'EtherODI' driver is shipped with most of this versions."
  echo ""
  echo "If you own a legal copy of Breadbox Ensemble, you can use the 'EtherPKT' driver"
  echo "provided with this skript. But be aware, that Breadbox does not give any support"
  echo "for this driver. So you are at your own."
  echo ""
  echo ""
  echo "4.  CONFIGURATIONS"
  echo ""
  echo "4.1 Configure this Script"
  echo ""
  echo " >> You must configure the user variables in this script which start with 'U_'."
  echo "    The script must be unlocked by setting the user variable 'U_SCRIPT_LOCKED'"
  echo "    to '0'."
  echo ""
  echo " >> Verify that the executable bits are set on this script. You can do this by"
  echo "    typing the following command in a shell:"
  echo ""
  echo "    chmod +x $(basename $0)"
  echo ""
  echo "4.2 Configure DOSEMU"
  echo ""
  echo " >> DOSEMU must be configured to use the following TAP adapter:"
  echo "    '$U_TAP_INTERFACE'"
  echo ""
  echo "    You do this by configuring the 'dosemu.conf' file like this:"
  echo "    '\$_pktdriver = (on)'"
  echo "    '\$_vnet = \"tap\"'"
  echo "    '\$_netdev = \"$U_TAP_INTERFACE\"'"
  echo ""
  echo "4.3 Configure DOS"
  echo ""
  echo "If you are using the 'EtherODI' driver, you must set up the ODI-to-Packet shim"
  echo "driver in DOSEMU. If you have Breadbox Ensemble and use the 'EtherPKT' driver,"
  echo "you must skip this section."
  echo ""
  echo " >> You must install the shim network drivers under DOSEMU:"
  echo "    'C:\NET\LSL.COM'"
  echo "    'C:\NET\PDETHER.EXE'"
  echo ""
  echo "    You must set up the configuration for the network drivers:"
  echo "    'C:\NET\NET.CFG'"
  echo ""
  echo "    You do this by configuring the 'NET.CFG' file like this:"
  echo "    Link Support"
  echo "        Max Boards 1"
  echo "        Max Stacks 3"
  echo "        Buffers 2 1514"
  echo "    Link Driver PDETHER"
  echo "        Int 60"
  echo "        Frame Ethernet_II"
  echo "        Protocol IP 800 Ethernet_II"
  echo "        Protocol ARP 806 Ethernet_II"
  echo "        Protocol RARP 8035 Ethernet_II"
  echo ""
  echo "    You must load this drivers in the 'AUTOEXEC.BAT':"
  echo "    C:"
  echo "    CD \NET"
  echo "    LSL.COM"
  echo "    LH PDETHER.EXE"
  echo ""
  echo "    On some installation, you receive the following message while loading"
  echo "    'PDETHER.EXE':"
  echo ""
  echo "        FATAL: Could not read MAC address from Packet Driver."
  echo "        Reason: Hardware not configured correctly."
  echo ""
  echo "    In this case, try to remove the 'LH' to load 'PDETHER.EXE' into convetional"
  echo "    memory instead into an UMB."
  echo ""
  echo "4.4 Configure PC/GEOS"
  echo ""
  echo " >> PC/GEOS must be configured to use the correct access point:"
  echo "    'accpnt'"
  echo ""
  echo "    You do this by configuring the 'GEOS.INI' file like this:"
  echo "    [accpnt]"
  echo "    prevID = 2"
  echo "    active1 = 2"
  echo "    contents = {"
  echo "    00020001"
  echo "    00010001"
  echo "    }"
  echo ""
  echo " >> PC/GEOS must be configured to use one of the following network drivers:"
  echo "    'EtherODI' or 'EtherPKT'"
  echo ""
  echo "    For 'EtherODI', you do this by configuring the 'GEOS.INI' file like this:"
  echo "    [tcpip]"
  echo "    driver = TCP/IP Driver"
  echo "    driverType = 0"
  echo "    link = ODI Ethernet Driver"
  echo "    ;linkPermName = ppp"
  echo "    linkDomain = EtherODI"
  echo "    port = 2"
  echo ""
  echo "    For 'EtherPKT', you do this by configuring the 'GEOS.INI' file like this:"
  echo "    [tcpip]"
  echo "    driver = TCP/IP Driver"
  echo "    driverType = 0"
  echo "    link = Packet Ethernet Driver"
  echo "    ;linkPermName = ppp"
  echo "    linkDomain = EtherPKT"
  echo "    port = 2"
  echo ""
  echo " >> PC/GEOS must be configured to use the following access point:"
  echo "    '$G_INI_SEC_AP2'"
  echo ""
  echo "    You do this by configuring the 'GEOS.INI' file like this:"
  echo "    [$G_INI_SEC_AP2]"
  echo "    name = Linux"
  echo "    netType = 1"
  echo "    default = 1"
  echo "    ipaddr = 192.168.0.200"
  echo "    ipgate = 192.168.0.254"
  echo "    ipmask = 255.255.255.0"
  echo "    dns1 = "
  echo "    dns2 = "
  echo "    phone = 0"
  echo "    promptSecret = 0"
  echo "    localDialingOptions = 0"
  echo ""
  echo " >> PC/GEOS must be configured to use a proxy:"
  echo "    '$G_INI_SEC_WEB'"
  echo ""
  echo "    [HTMLView]"
  echo "    ..."
  echo "    WWWProxy = {}"
  echo "    ..."
  echo ""
  echo "There are additional sections in the 'GEOS.INI' which must exist. The effective"
  echo "configuration of them can be retrieved from the original 'GEOS.INI' of your"
  echo "PC/GEOS version."
  echo ""
  echo " ** Hidden access point configuration:"
  echo "    [accesspoint0001]"
  echo "    name = hidden internet accesspoint"
  echo ""
  echo " ** TCP/IP socket configuration:"
  echo "    [socket]"
  echo "    domains = {"
  echo "    tcpip"
  echo "    tcpip_geos"
  echo "    loopback"
  echo "    }"
  echo "    driverCloseDelay = 0"
  echo "    LoadOnMsg = {"
  echo "    CDAB00000800060016000100loopback000700MBTRANS\\Socket Receive"
  echo "    CDAB00000A00060016000100tcpip_geos000700MBTRANS\\Socket Receive"
  echo "    CDAB00000500060016000100irlap000700MBTRANS\\Socket Receive"
  echo "    CDAB0C000500060016000100tcpip000700MBTRANS\\Socket Receive"
  echo "    }"
  echo ""
  echo " ** DNS configuration:"
  echo "    [resolver]"
  echo "    sbelt = {}"
  echo "    cacheSize = 100"
  echo "    errorCount = 0"
  echo ""
  return
}

#
# FUNCTION: doUp
#
# DESCRIPTION:
#  This function set up the networking.
#
doUp()
{
  checkScriptLocked
  checkInvocedByRoot
  checkWhichBinary "$G_WHICH_PATHS"
  checkLinuxBinaries "$G_REQ_LINUX_BINS"
  checkPackageBinaries "$G_REQ_BRIDGE_BINS" "bridge" "bridge-utils"
  checkPackageBinaries "$G_REQ_UML_BINS" "UML" "uml-utilities"
  checkPackageBinaries "$G_REQ_IPTABLES_BINS" "iptables" "iptables"
  checkProcFiles "$G_REQ_PROC_FILES"
  checkTapUser
  checkTapName
  checkGeosIniFile
  checkGeosActiveFile
  checkTapKernelModule
  checkBridgeKernelModule
  findActiveNetworkInterface
  findActiveIpAddress
  findFreeBridgeNetwork
  findDomainNameServers
  findHttpProxyServer
  loadTapKernelModule
  loadBridgeKernelModule
  enableIpForwarding
  setupNetworkBridge
  setupTapAdapter
  setupIpTables
  updateGeosIniFile
  printScriptConfig
  printRunningConfig
  return
}

#
# FUNCTION: doDown $1
#  $1		: If parameter equals '--complete', then disable the IP forwarding and
#         unload the kernel modules.
#
# DESCRIPTION:
#  This function shuts down the networking.
#
doDown()
{
  checkScriptLocked
  checkInvocedByRoot
  checkGeosActiveFile
  cleanupIpTables
  removeTapAdapter
  removeNetworkBridge
  if [ "$1" = "1" ] ;
  then
    checkTapKernelModule
    checkBridgeKernelModule
    disableIpForwarding
    unloadTapKernelModule
    unloadBridgeKernelModule
  fi
  return
}

#
# FUNCTION: doRestart
#
# DESCRIPTION:
#  This function restart the networking.
#
doRestart()
{
  # Shut down the networking
  checkScriptLocked
  checkInvocedByRoot
  checkGeosIniFile
  checkGeosActiveFile
  cleanupIpTables
  removeTapAdapter
  removeNetworkBridge
  # Start the networking
  checkTapKernelModule
  checkBridgeKernelModule
  findActiveNetworkInterface
  findActiveIpAddress
  findFreeBridgeNetwork
  findDomainNameServers
  findHttpProxyServer
  setupNetworkBridge
  setupTapAdapter
  setupIpTables
  updateGeosIniFile
  printScriptConfig
  printRunningConfig
  return
}

#
# FUNCTION: doReload
#
# DESCRIPTION:
#  This function reload the networking.
#
doReload()
{
  checkScriptLocked
  checkInvocedByRoot
  cleanupIpTables
  checkTapUser
  checkGeosIniFile
  checkTapKernelModule
  checkBridgeKernelModule
  findActiveNetworkInterface
  findActiveIpAddress
  findActiveBridgeNetwork
  findDomainNameServers
  findHttpProxyServer
  setupIpTables
  printScriptConfig
  printRunningConfig
  return
}

#
# FUNCTION: doStatus
#
# DESCRIPTION:
#  This function retrieves the status of the networking.
#
doStatus()
{
  checkTapUser
  checkGeosIniFile
  checkTapKernelModule
  checkBridgeKernelModule
  findActiveNetworkInterface
  findActiveIpAddress
  findActiveBridgeNetwork
  findDomainNameServers
  findHttpProxyServer
  printRunningConfig
  return
}


# ============================================================================
#  6.	Execute the selected function.
# ----------------------------------------------------------------------------

# Parse the command line arguments
L_ERR=0
for L_ARG in $*
do
  case $L_ARG in
  "-v" | "--version" )
    G_ARG_CMD="version"
  ;;
  "-c" | "--config" )
    G_ARG_CMD="config"
  ;;
  "-h" | "--help" )
    G_ARG_CMD="help"
  ;;
  "-u" | "--up" | "start" )
    G_ARG_CMD="start"
  ;;
  "-d" | "--down" | "stop" )
    G_ARG_CMD="stop"
  ;;
  "-r" | "--restart" | "restart" )
    G_ARG_CMD="restart"
  ;;
  "-l" | "--reload" | "reload" )
    G_ARG_CMD="reload"
  ;;
  "-s" | "--status" | "status" )
    G_ARG_CMD="status"
  ;;
  "--complete" )
    G_ARG_OPT_COMPLETE=1
  ;;
  "--force" )
    G_ARG_OPT_FORCE=1
  ;;
  "--verbose" )
    G_ARG_OPT_VERBOSE=1
  ;;
  * )
    L_ERR=1
  ;;
  esac
done
# Check for errors
if [ $L_ERR -ne 0 ] ;
then
  printScriptUsage
  echo "  ERROR - The script arguments are invalid !!!"
  echo "  VALUE : '$*'"
  echo ""
  handleError
fi
# Execute the command
case "$G_ARG_CMD" in
  "" )
    printScriptUsage
    exit 0
  ;;
  "version" )
    printScriptVersion
    exit 0
  ;;
  "config" )
    printScriptConfig
    exit 0
  ;;
  "help" )
    printScriptHelp|less
    exit 0
  ;;
  "start" )
    doUp
    echo ""
    echo "GEOS TAP Control has started successfully!"
    echo ""
    exit 0
  ;;
  "stop" )
    doDown "$G_ARG_OPT_COMPLETE"
    echo ""
    echo "GEOS TAP Control has removed successfully!"
    echo ""
    exit 0
  ;;
  "restart" )
    doRestart
    echo ""
    echo "GEOS TAP Control has restarted successfully!"
    echo ""
    exit 0
  ;;
  "reload" )
    doReload
    echo ""
    echo "GEOS TAP Control has reloaded successfully!"
    echo ""
    exit 0
  ;;
  "status" )
    doStatus
    echo ""
    echo "GEOS TAP Control has retrieved the status successfully!"
    echo ""
    exit 0
  ;;
esac

